	.include "macro.i"
	.include "cs8900a.i"


	.export main


	.import cs_packet_page
	.import cs_packet_data
	.import cs_rxtx_data
	.import cs_tx_cmd
	.import cs_tx_len

	.import cfg_ip
	.import cfg_mac


ptr	= $ae


	.bss

stack:		.res 1	; 2
len		= $57	; 9
pc_mac:		.res 6	; 4
pc_ip:		.res 4	; 6
ip_len		= $fd	; 11
ip_id:		.res 2	; 5
ip_cksum	= $fb	; 14
icmp_cksum:	.res 2	; 3
packet_len	= $ff	; 5

;udp_len:	.res 2
udp_dest_port:	.res 2
udp_src_port:	.res 2


	.segment "BUFFER"

	.align 256

packet_data:	.res $100


	.ifdef DEBUG

	.macro debug_printhex
	jsr printhex
	.endmacro

	.else

	.macro debug_printhex
	.endmacro

	.endif


	.code

; read data from the buffer, aborting to the main routine if the buffer is empty
read_word:
	lda len + 1
	bmi abort
	bne @gotdata
	lda len
	beq abort
@gotdata:
	lda len
	sec
	sbc #2
	sta len
	bcs :+
	dec len + 1
:	lda cs_rxtx_data
	.ifdef DEBUG
	pha
	debug_printhex
	lda cs_rxtx_data + 1
	pha
	debug_printhex
	pla
	tax
	pla
	.else
	ldx cs_rxtx_data + 1
	.endif
	rts


	.ifdef DEBUG
printhex:
	pha
	lsr
	lsr
	lsr
	lsr
	jsr @print
	pla
	pha
	and #$0f
	jsr @print
	pla
	rts

@print:
	ora #$30
	cmp #$3a
	bcc :+
	adc #6
:	jmp $ffd2
	.endif


purge:
	.ifdef DEBUG
	lda #2
	sta $d020
	lda #10
	sta 646
	.endif
:	jsr read_word			; read data until buffer is empty and read_word aborts
	jmp :-


abort:
	ldx stack
	txs
	; fall through to main

; wait for packets
main:
	tsx
	stx stack

	.ifdef DEBUG
	lda #$0d
	jsr $ffd2
	lda #14
	sta 646
	lda #14
	sta $d020
	.endif

	lda #$24			; check rx status
	sta cs_packet_page
	lda #$01
	sta cs_packet_page + 1

@wait:
	lda cs_packet_data + 1
	and #$0d
	beq @wait

	.ifdef DEBUG
	lda #7
	sta $d020
	.endif

	lda cs_rxtx_data + 1		; ignore status
	debug_printhex
	lda cs_rxtx_data
	debug_printhex
;	lda cs_rxtx_data		; ignore something else
;	debug_printhex
;	lda cs_rxtx_data + 1
;	debug_printhex

	lda cs_rxtx_data + 1		; read packet length
	sta len + 1
	debug_printhex
	lda cs_rxtx_data
	sta len
	debug_printhex

	.ifdef DEBUG
	inc 646
	.endif

	ldy #3				; skip dest mac (me or broadcast)
	jsr read_multiword

	.ifdef DEBUG
	dec 646
	.endif

	jsr read_pc_mac

	.ifdef DEBUG
	lda #7
	sta 646
	.endif

	jsr read_word
	cmp #$08
	bne @purge
	cpx #6
	beq arp
	cpx #0
	beq ip
@purge:
	jmp purge


arp:
	.ifdef DEBUG
	lda #5
	sta $d020
	lda #15
	sta 646
	.endif

	ldy #4				; discard first 6 bytes
	jsr read_multiword		; and read arp command

	cpx #1				; make sure it's a request
	bne @arp_purge

	.ifdef DEBUG
	inc 646
	.endif

	jsr read_pc_mac			; read senders mac address

	.ifdef DEBUG
	inc 646
	.endif

	jsr read_pc_ip			; read senders ip address

	.ifdef DEBUG
	inc 646
	inc 646
	.endif

	ldy #4				; discard target mac address
	jsr read_multiword		; and read first two target ip bytes

	cmp cfg_ip			; check if they're asking for my ip
	bne @arp_purge
	cpx cfg_ip + 1
	bne @arp_purge
	jsr read_word
	cmp cfg_ip + 2
	bne @arp_purge
	cpx cfg_ip + 3
	bne @arp_purge

	; here we should purge data in a friendly manner...

	ldy #6
	ldax #42
	jsr startreply			; start sending reply packet

	ldy #0				; send header
:	lda arp_reply_header,y
	iny
	ldx arp_reply_header,y
	jsr send_word
	iny
	cpy #8
	bne :-

	jsr send_my_mac
	jsr send_my_ip
	jsr send_pc_mac
	jsr send_pc_ip

	jmp main

@arp_purge:
	jmp purge


arp_reply_header:
	.dbyt $0001			; hw addr type (Ethernet)
	.dbyt $0800			; proto addr type (IPv4)
	.byte $06			; hw addr len
	.byte $04			; proto addr len
	.dbyt $0002			; reply


ip:
	.ifdef DEBUG
	lda #13
	sta $d020
	lda #1
	sta 646
	.endif

	ldy #2				; discard version, ihl, tos
	jsr read_multiword		; read len
	stx ip_len
	sta ip_len + 1
	lda ip_len
	ldx ip_len + 1
	sec				; subtract ip header length from packet length
	sbc #20
	bcs :+
	dex
:
	sta packet_len
	cpx #0				; restrict ip packet data to 256 bytes
	bne ip_purge

	jsr read_word
	sta ip_id
	stx ip_id + 1

	ldy #2				; discard id, flags, frag, ttl
	jsr read_multiword		; read protocol
	txa
	pha

	jsr read_word			; read checksum
	sta ip_cksum
	stx ip_cksum + 1

	jsr read_pc_ip

	ldy #2				; discard dest ip
	jsr read_multiword

	pla
	cmp #17
	beq udp
	cmp #1
	bne ip_purge
	jmp icmp
ip_purge:
	jmp purge


udp:
	.ifdef DEBUG
	lda #4
	sta 646
	.endif

	jsr read_word			; read udp ports
	stax udp_src_port
	jsr read_word
	stax udp_dest_port
	cmp #$0c			; check for port 3172
	bne ip_purge
	cpx #$64
	bne ip_purge

	jsr read_word			; get udp length
;	stax udp_len
	jsr read_word			; discard checksum

	.ifdef DEBUG
	lda #7
	sta 646
	.endif

	ldy #8
	jsr read_packet_data

	lda packet_data + 8
	cmp #$4c
	beq udp_load
	cmp #$53
	beq udp_save
	cmp #$52
	beq udp_run
	cmp #$4a
	beq udp_jump
	jmp main


; Load addrlo addrhi len data...
udp_load:
	jsr udp_ack
	dec $d020
	jsr bankinram
	lda packet_data + 10
	sta ptr
	lda packet_data + 11
	sta ptr + 1
	ldy #0
:	lda packet_data + 13,y
	sta (ptr),y
	iny
	cpy packet_data + 12
	bne :-
	jsr bankinio
	inc $d020
	jmp main


bankinram:
	sei
	lda #$34
	sta $01
	rts


bankinio:
	lda #$37
	sta $01
	cli
	rts


udp_save:
	inc $d020
	jmp *-3


udp_run:
	jsr udp_ack
	lda $ae
	sta $2d
	lda $af
	sta $2e
	jsr $a533
	jsr $a659
	jsr $a68e
	ldx #$f6
	txs
	jmp $a7ae


udp_jump:
	jsr udp_ack
	jmp (packet_data + 10)


icmp:
	.ifdef DEBUG
	lda #13
	sta 646
	.endif

	jsr read_word			; read type + code
	cmp #8
	bne icmp_purge

	jsr read_word			; read icmp_checksum
	sta icmp_cksum
	stx icmp_cksum + 1

	ldy #4
	jsr read_packet_data

	ldy #0
	ldax ip_len
	clc
	adc #14
	bcc :+
	inx
:	jsr startreply

	lda #$01
	jsr sendipheader

	lda #$00			; send reply packet
	tax
	jsr send_word

	ldax icmp_cksum			; adjust checksum
	clc
	adc #8
	bcc :+
	inx
:	jsr send_word

	ldy #4
:	cpy packet_len
	beq @done
	lda packet_data,y
	iny
	ldx packet_data,y
	jsr send_word
	iny
	bne :-
@done:
	jmp main
icmp_purge:
	jmp purge


udp_ack:
	lda #14 + 20 + 8 + 2		; eth + ip + udp + cmd + id
	ldx #0
	ldy packet_data + 8
	cpy #'S'			; save should also return data
	bne :+
	clc
	adc packet_data + 12
	bcc :+
	inx
:
	ldy #0
	jsr startreply

	lda #0				; reset checksum
	sta ip_cksum
	sta ip_cksum + 1
	sta ip_id
	sta ip_id + 1

	lda #$45			; add ipv4 and header length
	ldx #0
	jsr cksum_add

	lda #30				; add length of ip packet
	ldy packet_data + 8
	cpy #'S'
	bne :+
	clc
	adc packet_data + 12
:
	sta ip_len
	tax
	lda #0
	adc #0
	sta ip_len + 1
	jsr cksum_add

	; skip id and flags

	lda #$80			; add ttl and proto
	ldx #$11
	jsr cksum_add

	ldax cfg_ip
	jsr cksum_add
	ldax cfg_ip + 2
	jsr cksum_add

	ldax pc_ip
	jsr cksum_add
	ldax pc_ip + 2
	jsr cksum_add

	lda ip_cksum
	eor #$ff
	sta ip_cksum
	lda ip_cksum + 1
	eor #$ff
	sta ip_cksum + 1

	lda #$11
	jsr sendipheader

	ldax udp_dest_port		; flip src/dest port
	jsr send_word
	ldax udp_src_port
	jsr send_word

	lda ip_len			; send udp packet length
	sec
	sbc #20
	tax
	lda ip_len + 1
	sbc #0
	jsr send_word

	ldax #0				; skip udp checksum
	jsr send_word

	ldax packet_data + 8
	jsr send_word

	cmp #'S'
	bne @done

	lda packet_data + 12
	lsr
	bcc :+
	inc packet_data + 13
:
	ldy #0
:	lda packet_data + 13,y
	iny
	ldx packet_data + 13,y
	jsr send_word
	iny
	cpy packet_data + 12
	bne :-
@done:
	rts


sendipheader:
	pha

	lda #$45			; send ver, ihl, tos
	ldx #$00
	jsr send_word

	lda ip_len + 1			; send len
	ldx ip_len
	jsr send_word

	ldax ip_id			; send id
	jsr send_word

	ldax #0				; send flags
	jsr send_word

	pla				; send ttl and proto
	tax
	lda #$80
	jsr send_word

	ldax ip_cksum			; send checksum
	jsr send_word

	jsr send_my_ip
	jsr send_pc_ip

	rts


send_word:
	.ifdef DEBUG
	pha
	lda 646
	sta @save
	lda #12
	sta 646
	pla
	debug_printhex
	.endif

	sta cs_rxtx_data
	stx cs_rxtx_data + 1

	.ifdef DEBUG
	pha
	txa
	debug_printhex
@save = * + 1
	lda #$5e
	sta 646
	pla
	.endif

	rts


; length in A/X, type in Y
startreply:
	pha

	.ifdef DEBUG
	lda #'R'
	jsr $ffd2
	lda #'P'
	jsr $ffd2
	lda #'L'
	jsr $ffd2
	lda #'Y'
	jsr $ffd2
	.endif

	lda #$c9			; ask for buffer space
	sta cs_tx_cmd
	lda #0
	sta cs_tx_cmd + 1

	pla
	sta cs_tx_len			; set length
	stx cs_tx_len + 1

	lda #<pp_bus_status		; select bus status register
	sta cs_packet_page
	lda #>pp_bus_status
	sta cs_packet_page + 1

:	lda cs_packet_data + 1		; wait for space
	ldx cs_packet_data		; add timeout?
	lsr
	bcc :-

	jsr send_pc_mac
	jsr send_my_mac

	lda #$08
	sta cs_rxtx_data
	debug_printhex
	sty cs_rxtx_data + 1
	tya
	debug_printhex

	.ifdef DEBUG
	lda #'G'
	jsr $ffd2
	lda #'O'
	jsr $ffd2
	.endif

	rts


send_pc_mac:
	ldx #0				; send target mac
:	lda pc_mac,x
	sta cs_rxtx_data
	debug_printhex
	inx
	lda pc_mac,x
	sta cs_rxtx_data + 1
	debug_printhex
	inx
	cpx #6
	bne :-
	rts


send_my_mac:
	ldx #0				; send my mac
:	lda cfg_mac,x
	sta cs_rxtx_data
	debug_printhex
	inx
	lda cfg_mac,x
	sta cs_rxtx_data + 1
	debug_printhex
	inx
	cpx #6
	bne :-
	rts


send_pc_ip:
	ldx #0				; send target mac
:	lda pc_ip,x
	sta cs_rxtx_data
	debug_printhex
	inx
	lda pc_ip,x
	sta cs_rxtx_data + 1
	debug_printhex
	inx
	cpx #4
	bne :-
	rts


send_my_ip:
	ldx #0				; send my mac
:	lda cfg_ip,x
	sta cs_rxtx_data
	debug_printhex
	inx
	lda cfg_ip,x
	sta cs_rxtx_data + 1
	debug_printhex
	inx
	cpx #4
	bne :-
	rts


read_packet_data:
	lda packet_len			; round size up for odd sized payload
	lsr
	bcc :+
	inc packet_len
:	cpy packet_len
	beq @gotpacket
	jsr read_word
	sta packet_data,y
	txa
	iny
	sta packet_data,y
	iny
	bne :-
@gotpacket:
@flush:
	lda len + 1
	bmi @done
	bne @gotdata
	lda len
	beq @done
@gotdata:
	jsr read_word
	jmp @flush
@done:
	rts


read_multiword:
:	jsr read_word
	dey
	bne :-
	rts


read_pc_mac:
	ldy #0
:	jsr read_word
	sta pc_mac,y
	txa
	sta pc_mac + 1,y
	iny
	iny
	cpy #6
	bne :-
	rts


read_pc_ip:
	ldy #0
:	jsr read_word
	sta pc_ip,y
	txa
	sta pc_ip + 1,y
	iny
	iny
	cpy #4
	bne :-
	rts


cksum_add:
	pha
	txa
	clc
	adc ip_cksum + 1
	sta ip_cksum + 1
	pla
	adc ip_cksum
	sta ip_cksum
	bcc :+
	inc ip_cksum + 1
:	rts
