	.include "macro.i"
	.include "cs8900a.i"


	.export main


	.importzp rr_net_disable

	.import cs_packet_page
	.import cs_packet_data
	.import cs_rxtx_data
	.import cs_tx_cmd
	.import cs_tx_len

	.import cfg_ip
	.import cfg_mac


	.include "debug.i"


	.segment "BUFFER"
	.align 256
	.include "buffer.i"


ptr	= $2d


	.bss

;stack:		.res 1	; 2
len		= $57	; 9
;pc_mac:		.res 6	; 4
;pc_ip:		.res 4	; 6
;ip_len		= $fd	; 11
;ip_id:		.res 2	; 5
;ip_cksum	= $fb	; 14
;icmp_cksum:	.res 2	; 3
packet_len	= $ff	; 5

;udp_len:	.res 2
;udp_dest_port:	.res 2
;udp_src_port:	.res 2


	.code

arp_reply_header:
	.dbyt $0002			; reply
	.dbyt $0604			; addr len
	.dbyt $0800			; proto addr type (IPv4)
	.dbyt $0001			; hw addr type (Ethernet)



arp:
	ldx #3				; check if they're asking for us
@checkip:
	lda arp_tgt_ip,x
	cmp cfg_ip,x
	bne main
	dex
	bpl @checkip

	ldy #6
	ldax #42
	jsr startreply			; start sending reply packet

	ldy #6				; send header
:	lda arp_reply_header,y
	ldx arp_reply_header + 1,y
	jsr send_word
	dey
	dey
	bpl :-

	jsr send_my_mac
	jsr send_my_ip
	jsr send_pc_mac
	jsr send_pc_ip
	jmp main


purge:
@purge:
	ldx #128
:	lda cs_rxtx_data
	lda cs_rxtx_data + 1
	dex
	bne :-
	dey
	bne @purge

	;jmp main


; wait for packets
main:
	lda #$24			; check rx status
	sta cs_packet_page
	lda #$01
	sta cs_packet_page + 1

@wait:
	lda $c6
	beq :+
	dec $c6
	lda #rr_net_disable
	sta $de01
	rts
:	lda cs_packet_data + 1
	and #$0d
	beq @wait

	lda cs_rxtx_data + 1		; ignore status
	lda cs_rxtx_data

	lda cs_rxtx_data + 1		; read packet length
	sta len + 1
	lda cs_rxtx_data
	sta len
	lsr				; round up len
	bcc :+
	inc len
:
	ldx #0
@readpacket:
	lda cs_rxtx_data
	sta eth_packet,x
	inx
	lda cs_rxtx_data + 1
	sta eth_packet,x
	inx
	cpx len
	bne @readpacket

	ldy len + 1
	bne purge

	lda protocol
	cmp #8
	bne main
	lda protocol + 1
	cmp #6
	beq @arp
	cmp #0
	beq ip
	jmp main
@arp:
	jmp arp


ip:
	lda ip_proto
	cmp #1
	beq icmp
	cmp #17
	beq udp
	jmp main


icmp:
	lda icmp_type
	cmp #8
	bne main

	ldy #0
	lda ip_len + 1
	ldx ip_len
	clc
	adc #14
	bcc :+
	inx
:	jsr startreply

	lda #$01
	jsr sendipheader

	lda #$00			; send reply packet
	tax
	jsr send_word

	ldax icmp_cksum			; adjust checksum
	clc
	adc #8
	bcc :+
	inx
:	jsr send_word

	ldy #4
:	cpy packet_len
	beq @done
	lda icmp_packet,y
	iny
	ldx icmp_packet,y
	jsr send_word
	iny
	bne :-
@done:
icmp_purge:
	jmp main


udp:
	lda udp_dest_port
	cmp #$0c			; check for port 3172
	beq :+
@main:
	jmp main
:	lda udp_dest_port + 1
	cmp #$64
	bne @main

	lda udp_data
	cmp #$4c
	beq udp_load
	cmp #$53
	beq udp_save
	cmp #$52
	beq udp_run
	cmp #$4a
	beq udp_jump
	jmp main


; Load addrlo addrhi len data...
udp_load:
	jsr udp_ack
	dec $d020
	jsr bankinram
	lda udp_data + 2
	sta ptr
	lda udp_data + 3
	sta ptr + 1
	ldy #0
:	lda udp_data + 5,y
	sta (ptr),y
	iny
	cpy udp_data + 4
	bne :-
	jsr bankinio
	inc $d020
	jmp main


bankinram:
	sei
	lda #$34
	sta $01
	rts


bankinio:
	lda #$37
	sta $01
	cli
	rts


udp_save:
	inc $d020
	jmp *-3


udp_run:
	jsr udp_ack
	lda $ae
	sta $2d
	lda $af
	sta $2e
	jsr $a533
	jsr $a659
	jsr $a68e
	ldx #$f6
	txs
	jmp $a7ae


udp_jump:
	jsr udp_ack
	jmp (udp_data + 2)


udp_ack:
	lda #14 + 20 + 8 + 2		; eth + ip + udp + cmd + id
	ldx #0
	ldy udp_data - 8 + 8
	cpy #'S'			; save should also return data
	bne :+
	clc
	adc udp_data - 8 + 12
	bcc :+
	inx
:
	ldy #0
	jsr startreply

	lda #0				; reset checksum
	sta ip_cksum
	sta ip_cksum + 1
	sta ip_id
	sta ip_id + 1

	lda #$45			; add ipv4 and header length
	ldx #0
	jsr cksum_add

	lda #30				; add length of ip packet
	ldy udp_data - 8 + 8
	cpy #'S'
	bne :+
	clc
	adc udp_data - 8 + 12
:
	sta ip_len + 1
	tax
	lda #0
	sta ip_len
	jsr cksum_add

	; skip id and flags

	lda #$80			; add ttl and proto
	ldx #$11
	jsr cksum_add

	ldax cfg_ip
	jsr cksum_add
	ldax cfg_ip + 2
	jsr cksum_add

	ldax pc_ip
	jsr cksum_add
	ldax pc_ip + 2
	jsr cksum_add

	lda ip_cksum
	eor #$ff
	sta ip_cksum
	lda ip_cksum + 1
	eor #$ff
	sta ip_cksum + 1

	lda #$11
	jsr sendipheader

	ldax udp_dest_port		; flip src/dest port
	jsr send_word
	ldax udp_src_port
	jsr send_word

	lda ip_len + 1			; send udp packet length
	sec
	sbc #20
	tax
	lda ip_len
	sbc #0
	jsr send_word

	ldax #0				; skip udp checksum
	jsr send_word

	ldax udp_data - 8 + 8
	jsr send_word

	cmp #'S'
	bne @done

	lda udp_data - 8 + 12
	lsr
	bcc :+
	inc udp_data - 8 + 13
:
	ldy #0
:	lda udp_data - 8 + 13,y
	iny
	ldx udp_data - 8 + 13,y
	jsr send_word
	iny
	cpy udp_data - 8 + 12
	bne :-
@done:
	rts


sendipheader:
	pha

	lda #$45			; send ver, ihl, tos
	ldx #$00
	jsr send_word

	ldax ip_len			; send len
	jsr send_word

	ldax ip_id			; send id
	jsr send_word

	ldax #0				; send flags
	jsr send_word

	pla				; send ttl and proto
	tax
	lda #$80
	jsr send_word

	ldax ip_cksum			; send checksum
	jsr send_word

	jsr send_my_ip
	jsr send_pc_ip

	rts


send_word:
	.ifdef DEBUG
	pha
	lda 646
	sta @save
	lda #12
	sta 646
	pla
	debug_printhex
	.endif

	sta cs_rxtx_data
	stx cs_rxtx_data + 1

	.ifdef DEBUG
	pha
	txa
	debug_printhex
@save = * + 1
	lda #$5e
	sta 646
	pla
	.endif

	rts


; length in A/X, type in Y
startreply:
	pha

	.ifdef DEBUG
	lda #'R'
	jsr $ffd2
	lda #'P'
	jsr $ffd2
	lda #'L'
	jsr $ffd2
	lda #'Y'
	jsr $ffd2
	.endif

	lda #$c9			; ask for buffer space
	sta cs_tx_cmd
	lda #0
	sta cs_tx_cmd + 1

	pla
	sta cs_tx_len			; set length
	stx cs_tx_len + 1

	lda #<pp_bus_status		; select bus status register
	sta cs_packet_page
	lda #>pp_bus_status
	sta cs_packet_page + 1

:	lda cs_packet_data + 1		; wait for space
	ldx cs_packet_data		; add timeout?
	lsr
	bcc :-

	jsr send_pc_mac
	jsr send_my_mac

	lda #$08
	sta cs_rxtx_data
	debug_printhex
	sty cs_rxtx_data + 1
	debug_printhex

	.ifdef DEBUG
	lda #'G'
	jsr $ffd2
	lda #'O'
	jsr $ffd2
	.endif

	rts


send_pc_mac:
	ldx #0				; send target mac
:	lda pc_mac,x
	sta cs_rxtx_data
	debug_printhex
	inx
	lda pc_mac,x
	sta cs_rxtx_data + 1
	debug_printhex
	inx
	cpx #6
	bne :-
	rts


send_my_mac:
	ldx #0				; send my mac
:	lda cfg_mac,x
	sta cs_rxtx_data
	debug_printhex
	inx
	lda cfg_mac,x
	sta cs_rxtx_data + 1
	debug_printhex
	inx
	cpx #6
	bne :-
	rts


send_pc_ip:
	ldx #0				; send target mac
:	lda pc_ip,x
	sta cs_rxtx_data
	debug_printhex
	inx
	lda pc_ip,x
	sta cs_rxtx_data + 1
	debug_printhex
	inx
	cpx #4
	bne :-
	rts


send_my_ip:
	ldx #0				; send my mac
:	lda cfg_ip,x
	sta cs_rxtx_data
	debug_printhex
	inx
	lda cfg_ip,x
	sta cs_rxtx_data + 1
	debug_printhex
	inx
	cpx #4
	bne :-
	rts


cksum_add:
	pha
	txa
	clc
	adc ip_cksum + 1
	sta ip_cksum + 1
	pla
	adc ip_cksum
	sta ip_cksum
	bcc :+
	inc ip_cksum + 1
:	rts
